// Enable debug prints to serial monitor
#define MY_DEBUG 

// Enable and select radio type attached
#define MY_RADIO_NRF24
//#define MY_RADIO_RFM69

#define MY_NODE_ID       15

// Enable repeater functionality for this node
#define MY_REPEATER_FEATURE

#include <SPI.h>
#include <MySensors.h>
#include <Bounce2.h>
#include <TimeLib.h>

//#define TM1637
#define HT16K33
#ifdef TM1637
  #include <TM1637Display.h>
  #define PIN_ALARM_ARMED0   A3
  #define PIN_ALARM_ARMED1   A1
#endif
#ifdef HT16K33
  #include <Wire.h>
  #include <Adafruit_GFX.h>
  #include <Adafruit_LEDBackpack.h>
  #define SEG_A   0b00000001
  #define SEG_B   0b00000010
  #define SEG_C   0b00000100
  #define SEG_D   0b00001000
  #define SEG_E   0b00010000
  #define SEG_F   0b00100000
  #define SEG_G   0b01000000
  
  #define DISPLAY_ADDRESS   0x70
#endif

byte numbertable[] = {
	0x3F, /* 0 */
	0x06, /* 1 */
	0x5B, /* 2 */
	0x4F, /* 3 */
	0x66, /* 4 */
	0x6D, /* 5 */
	0x7D, /* 6 */
	0x07, /* 7 */
	0x7F, /* 8 */
	0x6F, /* 9 */
	0x77, /* a */
	0x7C, /* b */
	0x39, /* C */
	0x5E, /* d */
	0x79, /* E */
	0x71, /* F */
};

//#define PCBVERSION1
#define PCBVERSION2

#ifdef PCBVERSION1
  #define PIN_BUT_HOUR       8
  #define PIN_BUT_MINUTE     6
  #define PIN_BUT_SET_ALARM  4
  #define PIN_BUT_SNOOZE     2
  #define PIN_BUT_ALARM_OFF  7 
  #define RADIO_RELAY_OFF    LOW
  #define RADIO_RELAY_ON     HIGH
#endif
#ifdef PCBVERSION2 
  #define PIN_BUT_HOUR       2
  #define PIN_BUT_ALARM_OFF  4   
  #define PIN_BUT_MINUTE     6
  #define PIN_BUT_SET_ALARM  7
  #define PIN_BUT_SNOOZE     8
  #define RADIO_RELAY_OFF    HIGH
  #define RADIO_RELAY_ON     LOW
#endif


#define PIN_LIGHT_RED       3
#define PIN_LIGHT_GREEN     5

#define PIN_LDR            (byte)A0
#define PIN_RADIO          (byte)A2

// Snooze delay is 8 minutes
#define SNOOZE_DELAY 8

// Start lighting 30 minutes before alarm
#define WAKEUP_LIGHT_TIME 30

// Enter here the node and sensor if for which the temperature needs to be requested
#define SENSORNODEID       30
byte    tempsensors_ids[]  = {5, 4, 16};
byte    tempsensors_cnt    = 0;
#define MAXTEMPIDS        (int)((sizeof(tempsensors_ids)/sizeof(tempsensors_cnt)));

// Module connection pins (Digital Pins)
#ifdef TM1637
  #define CLK A5
  #define DIO A4
  TM1637Display display(CLK, DIO);
#endif
#ifdef HT16K33
   Adafruit_7segment display = Adafruit_7segment();
#endif
boolean timeReceived = false;
byte    time_data[5] = { SEG_G, 0x80 | SEG_G, SEG_G, SEG_G, 0x00 };
byte    auto_data[5] = { SEG_A | SEG_B | SEG_C | SEG_E | SEG_F | SEG_G,   // A
                         SEG_C | SEG_D | SEG_E,                           // u
                         SEG_D | SEG_E | SEG_F | SEG_G,                   // t
                         SEG_C | SEG_D | SEG_E | SEG_G,                   // o
                         0 };                 
byte    off_data[5] =  { SEG_A | SEG_B | SEG_C | SEG_E | SEG_F | SEG_G,   // A
                         SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F,   // O
                         SEG_A | SEG_G | SEG_E | SEG_F,                   // F
                         SEG_A | SEG_G | SEG_E | SEG_F,                   // F
                         0 };
byte    alon_data[5] = { SEG_A | SEG_B | SEG_C | SEG_E | SEG_F | SEG_G,   // A
                         SEG_D | SEG_E | SEG_F,                           // L
                         SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F,   // O
                         SEG_A | SEG_B | SEG_C | SEG_E | SEG_F,           // N
                         };        
byte    dark_data[5] =  {SEG_B | SEG_C | SEG_D | SEG_E | SEG_G,           // d
                         SEG_A | SEG_B | SEG_C | SEG_E | SEG_F | SEG_G,   // A
                         SEG_E | SEG_G,                                   // r
                         SEG_B | SEG_C | SEG_E | SEG_F | SEG_G,           // k
                         0 };
byte    light_data[5] = {SEG_F | SEG_E | SEG_C,                           // L i
                         SEG_A | SEG_B | SEG_C | SEG_D | SEG_F | SEG_G,   // g
                         SEG_C | SEG_E | SEG_F | SEG_G,                   // h
                         SEG_D | SEG_E | SEG_F | SEG_G,                   // t
                         0 }; 

byte    thisMinute       = 99;
byte    thisSecond       = 99;
byte    thisHour         = 99;

byte    snoozeHour       = 00;
byte    snoozeMinute     = 00;

#define NUM_OF_ALARMS    2
byte    alarmHour[NUM_OF_ALARMS]    = { 00, 00 };
byte    alarmMinute[NUM_OF_ALARMS]  = { 00, 00 };

byte    alarm_data[]     = { 0, 0, 0, 0, 0 };
byte    alarmArmed       = 1;
byte    temp_data[]      = { SEG_G, 0x80 | SEG_G, SEG_G, 0,  0x10 };
byte    manTriggerCnt    = 0;
byte    man_data[]       = { 0, 0, 0, 0, 0 };

bool    useLight         = true;

#define LIGHT_OFF          false
#define LIGHT_ON           true
#define NO_ACK             false
#define SEND_ACK           true
#define RENDER_COLON       true
#define USE_DATA_4_COLON   false
#define RADIO_OFF          false
#define RADIO_ON           true
#define RENDER_VALUE       -1
#define MAX_GREEN          128
#define MAX_RED            255

#define NODE_ID    15
#define NODE_RADIO  0
#define NODE_ALARM  1
#define NODE_LIGHT  2

// NRFRF24L01 radio driver (set low transmit power by default) 
// --- MyTransportNRF24 radio(RF24_CE_PIN, RF24_CS_PIN, RF24_PA_LEVEL_GW);  
//MyTransportRFM69 radio;
// Message signing driver (none default)
//MySigningNone signer;
// Select AtMega328 hardware profile
// --- MyHwATMega328 hw;
// Construct MySensors library
// --- MySensor gw(radio, hw);

MyMessage msgRadioOn     (NODE_RADIO, V_STATUS);
MyMessage msgAlarmArmed  (NODE_ALARM, V_ARMED);
MyMessage msgAlarmTripped(NODE_ALARM, V_TRIPPED);
MyMessage msgLightLevel  (NODE_LIGHT, V_PERCENTAGE);
MyMessage msgLightStatus (NODE_LIGHT, V_STATUS);

enum EEPROM_ADRES {
  EEPROM_ALARM_HOUR0 = 0,
  EEPROM_ALARM_MINUTE0,
  EEPROM_ALARM_ARMED,
  EEPROM_ALARM_HOUR1,
  EEPROM_ALARM_MINUTE1,
  EEPROM_ALARM_USE_LIGHT
};

enum STATES {
   STATE_TIME = 0,
   STATE_SET_ALARM,
   STATE_ALARM_ON,
   STATE_ALARM_UPCOMING,
   STATE_SNOOZING,
   STATE_PLAY_MUSIC,
   STATE_SHOW_TEMPERATUE,
   STATE_MANUAL_TRIGGER
} STATE_TYPE;
#define STATE_LAST_STATE STATE_MANUAL_TRIGGER

int  state     = STATE_TIME;
int  lastState = STATE_TIME;

// Button debouncers
Bounce setHourButton   = Bounce();
Bounce setMinuteButton = Bounce();
Bounce setAlarmButton  = Bounce();
Bounce snoozeButton    = Bounce();
Bounce alarmOffButton  = Bounce();


void before()
{
#ifdef HT16K33
   // Setup the display.
   display.begin(DISPLAY_ADDRESS);
#endif

   // Set up set LDR input
   pinMode(PIN_LDR, INPUT);
   updateDisplayLevel();   
   setSegments(time_data, USE_DATA_4_COLON);
}

void setup()  
{   
   // Read EEPROM settings   
   alarmHour[0]   = loadState(EEPROM_ALARM_HOUR0);  // Auto mode > 24 
   alarmMinute[0] = loadState(EEPROM_ALARM_MINUTE0) % 60;
   alarmHour[1]   = loadState(EEPROM_ALARM_HOUR1);
   alarmMinute[1] = loadState(EEPROM_ALARM_MINUTE1) % 60;
   alarmArmed     = loadState(EEPROM_ALARM_ARMED)   % (NUM_OF_ALARMS + 1); 
   useLight       = (loadState(EEPROM_ALARM_USE_LIGHT) > 0); 
   
   Serial.print(F("Alarm time: "));
   if (alarmHour[alarmArmed % NUM_OF_ALARMS] > 24) { 
       Serial.print(F("Auto (external trigger only)"));
   } else {
       Serial.print(alarmHour[alarmArmed % NUM_OF_ALARMS]);
       Serial.print(":");
       Serial.print(alarmMinute[alarmArmed % NUM_OF_ALARMS]);
   }
   Serial.println(alarmArmed > 0 ? " ARM" : " byp");
   Serial.println(useLight ? F("Lght") : F("Dark"));
   
   // Set up light level pin to 0
   pinMode(PIN_LIGHT_RED, OUTPUT);
   pinMode(PIN_LIGHT_GREEN, OUTPUT);
   setWakeUpLight(LIGHT_OFF);
   
   // Then set the radio pin in output mode and make sure it is silent
   pinMode(PIN_RADIO, OUTPUT);   
   digitalWrite(PIN_RADIO, RADIO_RELAY_OFF);
   
#ifdef TM1637   
   // Then set the alarm armed pin in output mode and make sure it is set to the last known state
   pinMode(PIN_ALARM_ARMED0, OUTPUT);   
   digitalWrite(PIN_ALARM_ARMED0, alarmArmed == 1 ? HIGH : LOW);

   // Then set the alarm armed pin in output mode and make sure it is set to the last known state
   pinMode(PIN_ALARM_ARMED1, OUTPUT);   
   digitalWrite(PIN_ALARM_ARMED1, alarmArmed == 2 ? HIGH : LOW);
#endif 

   // Set up Hour button
   pinMode(PIN_BUT_HOUR, INPUT_PULLUP);
   setHourButton.attach(PIN_BUT_HOUR);
   setHourButton.interval(5);

   // Set up Minute button
   pinMode(PIN_BUT_MINUTE, INPUT_PULLUP);
   setMinuteButton.attach(PIN_BUT_MINUTE);
   setMinuteButton.interval(5);

   // Set up set Alarm time button
   pinMode(PIN_BUT_SET_ALARM, INPUT_PULLUP);
   setAlarmButton.attach(PIN_BUT_SET_ALARM);
   setAlarmButton.interval(5);
   
   // Set up Snooze button
   pinMode(PIN_BUT_SNOOZE, INPUT_PULLUP);
   snoozeButton.attach(PIN_BUT_SNOOZE);
   snoozeButton.interval(5);
   
   // Set up Alarm off button
   pinMode(PIN_BUT_ALARM_OFF, INPUT_PULLUP);
   alarmOffButton.attach(PIN_BUT_ALARM_OFF);
   alarmOffButton.interval(5);
   
   // Send Celsius or Fahrenheit sign (C or F)
   temp_data[3] = (getControllerConfig().isMetric ? numbertable[0xC] : numbertable[0xF]); 
   wait(100);
   requestTime();
   unsigned int lastRequest = millis();
   while (!timeReceived) {
      wait(100);
      if ((lastRequest + 5000) < millis()) {
          requestTime(); 
#ifdef TM1637
          time_data[1] ^= 0x80;
#endif
#ifdef HT16K33
          time_data[4] ^= 0x02;
#endif
          setSegments(time_data, USE_DATA_4_COLON);
          lastRequest = millis();
      }
   }
   send(msgAlarmTripped.set("0"), true);

   Serial.println(F("Init"));
}


void presentation ()
{
   // Send the sketch version information to the gateway and Controller
   sendSketchInfo("ClockRadio", "1.0");
   wait(100);
   
   // Register all sensors to gw (they will be created as child devices)
   present(NODE_RADIO, S_LIGHT, "Clock Radio");        wait(100);
   present(NODE_ALARM, S_MOTION, "Clock Radio Alarm"); wait(100);
   present(NODE_LIGHT, S_DIMMER, "Wakeup light");      wait(100);
}

void loop() 
{
   /* ===================== Check time based activities ======================== */
   // update current time
   if (thisSecond != second()) {
      thisSecond = second();
      thisHour   = hour();

      if (thisMinute != minute())
      {   // Minute changed so update clock
         thisMinute = minute();
         timeToArray(time_data, thisHour, thisMinute);
         Serial.print(F("Time upd: "));
         Serial.print(thisHour);
         Serial.print(":");
         Serial.println(thisMinute);
         
         switch (state) {
             case STATE_SHOW_TEMPERATUE:
             case STATE_MANUAL_TRIGGER:
             case STATE_SET_ALARM:
                break;
              
             case STATE_PLAY_MUSIC:
                if (snoozeHour == thisHour && snoozeMinute == thisMinute) {
                   Serial.println(F("Snz tm over stop m"));
                   setState(STATE_TIME);
                   switchRadio(RADIO_OFF, SEND_ACK);
                }
                setSegments(time_data, RENDER_COLON);
                break;
             
             case STATE_SNOOZING:
                if (snoozeHour == thisHour && snoozeMinute == thisMinute) {
                   Serial.println(F("Snz tm over rdo on"));
                   setState(STATE_ALARM_ON);
                   switchRadio(RADIO_ON, SEND_ACK);
                   
                   // Stop Alarm after 1 hour
                   snoozeHour   = (thisHour + 1) % 24;
                   snoozeMinute = thisMinute;
                }
                setSegments(time_data, RENDER_COLON);
                break;
             
             case STATE_TIME:
                 if (alarmArmed > 0 && useLight)
                 {    // Calculate the wake time in total minutes (make sure we cover mid night-passing)
                	 int wakeTime = (alarmHour[alarmArmed % NUM_OF_ALARMS] * 60) + alarmMinute[alarmArmed % NUM_OF_ALARMS] +
                		((alarmHour[alarmArmed % NUM_OF_ALARMS] == 00 && alarmMinute[alarmArmed % NUM_OF_ALARMS] < WAKEUP_LIGHT_TIME) ? (24*60) : 0);
                	  if ((wakeTime - ((thisHour * 60) +  thisMinute)) == WAKEUP_LIGHT_TIME) {
                    	  setState(STATE_ALARM_UPCOMING);
                          snoozeHour   = alarmHour[alarmArmed % NUM_OF_ALARMS];
						  snoozeMinute = alarmMinute[alarmArmed % NUM_OF_ALARMS];
                      }
                 }
                 /* no break */
             case STATE_ALARM_UPCOMING:
                if ((alarmHour[alarmArmed % NUM_OF_ALARMS] == thisHour && alarmMinute[alarmArmed % NUM_OF_ALARMS] == thisMinute) ||
				    (state == STATE_ALARM_UPCOMING && snoozeHour == thisHour && snoozeMinute == thisMinute))  {
                   // OKE it is time to sound theALARM 
                   if (alarmArmed > 0) { // If we're armed of course
                       Serial.println(F("-- ALARM --"));
                       setState(STATE_ALARM_ON);
                       send(msgAlarmTripped.set("1"), true);
                       wait(100);
                       switchRadio(RADIO_ON, SEND_ACK);
                       if (useLight) {
                           setWakeUpLight(LIGHT_ON);
                       } 
                       // Stop Alarm after 1 hour
                       snoozeHour   = (thisHour + 1) % 24;
                       snoozeMinute = thisMinute;
                   } else {
                	   setState(STATE_TIME);
                   }
                }
                setSegments(time_data, RENDER_COLON);
                break;
                
            case STATE_ALARM_ON:
                if (snoozeHour == thisHour && snoozeMinute == thisMinute) {
                   // Alarm was on for one hour, assume false alarm so stop it
                   Serial.println(F("False alrm stp"));
                   setState(STATE_TIME);
                   send(msgAlarmTripped.set("0"), true);
                   wait(100);
                   switchRadio(RADIO_OFF, SEND_ACK);
                   setWakeUpLight(LIGHT_OFF);
                }
                setSegments(time_data, RENDER_COLON);
                break;
         }
      } else {
          switch (state) {
             case STATE_MANUAL_TRIGGER: // Manual trigger button is active
                if (manTriggerCnt) {
                   manTriggerCnt--;
                   if (manTriggerCnt == 0) {
                       // After four seconds still pressing -> send alarm trigger event to GW, 
                       // can be used to switch on coffee machine with out the need for an alarm
                	   // Only show first 2 digits (AL) on ack ON will be shown aswell
                       man_data[0] = alon_data[0];
                       man_data[1] = alon_data[1];
                       send(msgAlarmTripped.set("1"), true);
                   } else {
                	   // Show extra - digit, to let the user know his button press is noticed
                       man_data[manTriggerCnt-1] = SEG_G; 
                   }
                   setSegments(man_data, USE_DATA_4_COLON);
                }
                break;

             case STATE_SET_ALARM:
             case STATE_SHOW_TEMPERATUE:
                break;
             
             case STATE_TIME:
             case STATE_ALARM_UPCOMING:
             case STATE_ALARM_ON:
             case STATE_SNOOZING:
             case STATE_PLAY_MUSIC:
                setSegments(time_data, RENDER_COLON);
                break;
          }
      }
      
      // Sync system time each 10 minutes and 30 seconds (so we do not have to change the time displayed and check for any alarm/state change)
      // Assuming we do not drift 30 second or more in 10 minutes
      if (thisSecond == 30 && (thisMinute % 10) == 0) {
			  timeReceived = false;
	  }

      // Render each second the wake up light level (if ALARM in upcoming)
      if (state == STATE_ALARM_UPCOMING) { 
          wakeupLight(RENDER_VALUE);
      }

      // if temperature not yet has been received give it a retry each second
      if (state == STATE_SHOW_TEMPERATUE && temp_data[2] == SEG_G) {
   		  request(tempsensors_ids[tempsensors_cnt], V_TEMP, SENSORNODEID);
      } else {
    	  // If time was invalidated re-send requestTime() every second until a fresh time was received
          if (!timeReceived) {
   	      	  requestTime();
          }
      }
    
      // Render LDR level to adjust LED display brightness, only once a second
      updateDisplayLevel();
   }
   
   /* ===================== Now check button activities ======================== */

   // Check of Set Alarm button was changed 
   if (setAlarmButton.update()) {
       if (setAlarmButton.read()) { // now button released detected save new Alarm time setting
           Serial.println(F("Leave set ALARM tm"));
           setState(lastState);
           // Save new Alarm time in EEPROM
           saveState(EEPROM_ALARM_HOUR0,     alarmHour[0]);
           saveState(EEPROM_ALARM_MINUTE0,   alarmMinute[0]);
           saveState(EEPROM_ALARM_HOUR1,     alarmHour[1]);
           saveState(EEPROM_ALARM_MINUTE1,   alarmMinute[1]);
           saveState(EEPROM_ALARM_USE_LIGHT, useLight); 

           Serial.print(F("Alarm time: "));
           Serial.print(alarmHour[alarmArmed % NUM_OF_ALARMS]);
           Serial.print(":");
           Serial.println(alarmMinute[alarmArmed % NUM_OF_ALARMS]);
           // Show current time again
           setSegments(time_data, RENDER_COLON);
       } else {
           Serial.println(F("Enter set ALARM"));
           // Remember the current state to return to when new alarm-time is set
           lastState = state;
           setState(STATE_SET_ALARM);

           if (alarmArmed > 0) { // We're in auto mode (external trigger only)
              if (alarmHour[alarmArmed % NUM_OF_ALARMS] > 23) {
                  setSegments(auto_data, USE_DATA_4_COLON);
              } else { /// Show Alarm time
                  timeToArray(alarm_data, alarmHour[alarmArmed % NUM_OF_ALARMS], alarmMinute[alarmArmed % NUM_OF_ALARMS]);
                  setSegments(alarm_data, USE_DATA_4_COLON);
              }
           } else {  // No Alarm active show "off"
              setSegments(off_data, USE_DATA_4_COLON);
           }
       }
   }
   if (setHourButton.update()) {
       switch (state) {
           case STATE_MANUAL_TRIGGER:
               break;

           case STATE_SET_ALARM: 
               if (alarmArmed > 0 && setHourButton.read()) {
                  // Set HOUR pressed
                  if (alarmHour[alarmArmed % NUM_OF_ALARMS] == 0 && alarmMinute[alarmArmed % NUM_OF_ALARMS] == 0) { // Do we enter auto mode? 
                      alarmHour[alarmArmed % NUM_OF_ALARMS] = 48;
                      setSegments(auto_data, USE_DATA_4_COLON);
                  } else {
                      alarmHour[alarmArmed % NUM_OF_ALARMS] = (alarmHour[alarmArmed % NUM_OF_ALARMS] + 1) % 24;
                      // Show new Alarm time
                      timeToArray(alarm_data, alarmHour[alarmArmed % NUM_OF_ALARMS], alarmMinute[alarmArmed % NUM_OF_ALARMS]);
                      setSegments(alarm_data, USE_DATA_4_COLON);
                  }
               }
               break;
             
          case STATE_SHOW_TEMPERATUE:
               if (setHourButton.read()) {
                  Serial.println(F("Leave show tmp"));
                  // Switch back to previous state
                  setState(lastState);
                  setSegments(time_data, RENDER_COLON);
               }
               break;
               
          case STATE_SNOOZING:
          case STATE_ALARM_ON:
          case STATE_PLAY_MUSIC:
          case STATE_TIME:
          case STATE_ALARM_UPCOMING:
               if (!setHourButton.read()) {
                  Serial.println(F("Enter show temp"));
                  tempsensors_cnt  = (tempsensors_cnt + 1) % MAXTEMPIDS;
                  request(tempsensors_ids[tempsensors_cnt], V_TEMP, SENSORNODEID);
                  // Remember the current state to return to when temperature display is over
                  lastState = state;
                  setState(STATE_SHOW_TEMPERATUE);
                  for (byte i = 0; i < 3; i++ ) {
                      temp_data[i] = SEG_G;
                  }
                  setSegments(temp_data, USE_DATA_4_COLON);
               }
               break;
       }
   }
   if (setMinuteButton.update()) {
       switch (state) {
          case STATE_SET_ALARM:
             if (alarmArmed > 0 && setMinuteButton.read()) {
                // Set MINUTE pressed
                if (alarmHour[alarmArmed % NUM_OF_ALARMS] == 0 && alarmMinute[alarmArmed % NUM_OF_ALARMS] == 0) { // Do we enter auto mode? 
                      alarmHour[alarmArmed % NUM_OF_ALARMS] = 48;
                      setSegments(auto_data, USE_DATA_4_COLON);
                } else {
                      alarmHour[alarmArmed % NUM_OF_ALARMS] %= 24; // Make sure we leave auto mode
                      alarmMinute[alarmArmed % NUM_OF_ALARMS] = (alarmMinute[alarmArmed % NUM_OF_ALARMS] + 1) % 60;
                      // Show new Alarm time
                      timeToArray(alarm_data, alarmHour[alarmArmed % NUM_OF_ALARMS], alarmMinute[alarmArmed % NUM_OF_ALARMS]);
                      setSegments(alarm_data, USE_DATA_4_COLON);
                }
             }
             break;
          
          case STATE_MANUAL_TRIGGER:
             if (setMinuteButton.read()) {
                 Serial.println(F("Leave man-trig"));
                 // Return to previous state 
                 setState(lastState);
                 if (manTriggerCnt == 0) {
                     // Manual trigger was set so clear now
                     send(msgAlarmTripped.set("0"), true);
                 }
             }
             break;
             
          case STATE_PLAY_MUSIC:
          case STATE_TIME:
          case STATE_ALARM_UPCOMING:
             if (!setMinuteButton.read()) { 
                 Serial.println(F("Enter man-trig"));
                 // Remember the current state to return to when finished signaling the gateway
                 lastState = state;
                 setState(STATE_MANUAL_TRIGGER);
                 manTriggerCnt = 4;
                 for (byte i = 0; i < 3; i++)  {
                     man_data[i] = 0; 
                 }
                 man_data[3] = SEG_G;
                 setSegments(man_data, USE_DATA_4_COLON);
             }
             break;

          case STATE_SHOW_TEMPERATUE:
          case STATE_SNOOZING:
          case STATE_ALARM_ON:
        	 break;
       }
   }
   if (snoozeButton.update()) {
      bool bOff = (snoozeButton.read() ? false : true); // this is an inverted button
      switch (state) {
          case STATE_ALARM_ON:
              if (bOff) {
                  Serial.println(F("Snz Alarm for 8m"));
                  setState(STATE_SNOOZING);
                  switchRadio(RADIO_OFF, SEND_ACK);
                  snoozeMinute = (thisMinute + SNOOZE_DELAY) % 60;
                  snoozeHour   = (snoozeMinute > thisMinute ? thisHour : (thisHour + 1) % 24);
              }
              break;
 
          case STATE_PLAY_MUSIC:
              if (bOff) {
                  Serial.println(F("Snz in music, rd off"));
                  setState(STATE_TIME);
                  switchRadio(RADIO_OFF, SEND_ACK);
              }
              break;
              
          case STATE_TIME:
          case STATE_ALARM_UPCOMING:
              if (bOff) {
                  Serial.println(F("Snz in resttime music 1h"));
                  setState(STATE_PLAY_MUSIC);
                  snoozeMinute = thisMinute;
                  snoozeHour   = (thisHour + 1) % 24;
                  switchRadio(RADIO_ON, SEND_ACK);
              }
              break; 
              
           case STATE_SET_ALARM:
              if (!bOff) {
                  Serial.println(F("Sw light on"));
                  setSegments(light_data, USE_DATA_4_COLON);
                  useLight = true;
              } else {
                  setSegments(((alarmArmed > 0) ? ((alarmHour[alarmArmed % NUM_OF_ALARMS] > 24) ? auto_data :  alarm_data) : off_data) , USE_DATA_4_COLON);
              }
              break; 

           case STATE_SHOW_TEMPERATUE:
           case STATE_SNOOZING:
           case STATE_MANUAL_TRIGGER:
              break;
      }
   }
   if (alarmOffButton.update()) {
      bool bOff = alarmOffButton.read();
      switch (state) {
          case STATE_ALARM_ON: 
          case STATE_SNOOZING:
              if (bOff) {
                  Serial.println(F("Wr awake stp alrm"));
                  setState(STATE_TIME);
                  send(msgAlarmTripped.set("0"), true);
                  wait(100);
                  switchRadio(RADIO_OFF, SEND_ACK);
                  setWakeUpLight(LIGHT_OFF);
              }
              break; 
             
          case STATE_PLAY_MUSIC:
          case STATE_TIME:
              if (bOff) {
                  Serial.println(F("Alrm tgl press"));
                  send(msgAlarmArmed.set(alarmArmed < NUM_OF_ALARMS), true);
              }
              break;
          
          case STATE_ALARM_UPCOMING:
              if (bOff) {
                  Serial.println(F("Alrm press dim lght"));
                  setWakeUpLight(LIGHT_OFF);
                  setState(STATE_TIME);
              }
              break;
              
          case STATE_SET_ALARM:
              if (!bOff) {
                  Serial.println(F("Sw light off"));
                  setSegments(dark_data, USE_DATA_4_COLON);
                  // Make sure the lights are off
                  setWakeUpLight(LIGHT_OFF);
                  useLight = false;
              } else {
                  setSegments(((alarmArmed > 0) ? ((alarmHour[alarmArmed % NUM_OF_ALARMS] > 24) ? auto_data :  alarm_data) : off_data) , USE_DATA_4_COLON);
              }
              break; 

          case STATE_SHOW_TEMPERATUE:
          case STATE_MANUAL_TRIGGER:
              break;
      }
   }
   wait(10);
}
 
void updateDisplayLevel() {
   static byte desiredLevel = 0;
   static byte currentLevel = 0x06;

#ifdef TM1637
   int newLevel = map(analogRead(PIN_LDR), 500, 1000, 0x08, 0x0F);
   newLevel = (newLevel < 8 ? 8 : newLevel);
   newLevel = (newLevel > 0x0F ? 0x0F : newLevel);
#endif
#ifdef HT16K33
   int ldrraw = analogRead(PIN_LDR);
   Serial.println(ldrraw);
   //int newLevel = (int)fscale( 10, 800, 0, 0x0F, ldrraw, -4);
   int newLevel = (int)fscale( 600, 1000, 0, 0x0F, ldrraw, -4);
#endif
   if (newLevel != desiredLevel) {
       // Wait at least 1 seconds before new light level is set
       desiredLevel = newLevel;
   } else {
       if (desiredLevel != currentLevel) {
          // New level for 2 seconds stable move newlevel one step to desired level
          currentLevel = (desiredLevel > currentLevel ? currentLevel + 1 : currentLevel - 1); 
          Serial.print("Light level: ");
          Serial.println(currentLevel);
          display.setBrightness(currentLevel); 
       }
   }
}

 
void timeToArray (byte times[], byte bHour, byte bMin) {
    // Minute changed so update clock
    times[0] = (bHour > 9 ? numbertable[bHour / 10] : 0);
    times[1] = numbertable[bHour % 10];
    times[2] = (bMin > 9 ? numbertable[bMin / 10] : numbertable[0]);
    times[3] = numbertable[bMin  % 10]; 
}

void tempToArray (byte digits[], float temp, byte id) {
    // Calculate temp to digits
    if (temp < 0) {
        digits[0] = SEG_G;
        temp *= -1;
        if (temp >= 10) {
            digits[1] = numbertable[(int)(temp / 10)];
            digits[2] = numbertable[((int)temp) % 10];
            digits[4] = 0x10;
        } else {
            digits[1] = numbertable[((int)temp) % 10] | 0x80;
            digits[2] = numbertable[(int)(temp * 10) % 10];
            digits[4] = 0x10 | 0x02;
        }  
    } else {
          digits[0] = (temp > 10 ? numbertable[(int)(temp / 10)] : 0);
          digits[1] = numbertable[((int)temp) % 10] | 0x80;
          digits[2] = numbertable[(int)(temp * 10) % 10];
          //digits[4] = 0x10 | 0x02 | ((id % 0x01) == 0  ? 0x00 : 0x04 ) | ((id % 0x02) == 0  ? 0x00 : 0x08 );
          digits[4] = 0x10 | 0x02 | (id << 2);
    }
}

void setSegments(byte data[5], bool renderColon) {
#ifdef TM1637   
   // Then set the alarm armed pin in output mode and make sure it is set to the last known state
   pinMode(PIN_ALARM_ARMED0, OUTPUT);   
   digitalWrite(PIN_ALARM_ARMED0, alarmArmed == 1 ? HIGH : LOW);

   // Then set the alarm armed pin in output mode and make sure it is set to the last known state
   pinMode(PIN_ALARM_ARMED1, OUTPUT);   
   digitalWrite(PIN_ALARM_ARMED1, alarmArmed == 2 ? HIGH : LOW);
   
   if (renderColon) {
     if ((second() % 2) == 0) {
        data[1] = data[1] & 0x7f;
     } else {
        data[1] = data[1] | 0x80;
     }
   }
   display.setSegments(data);
#endif 
#ifdef HT16K33
    display.writeDigitRaw(0, data[0]);
    display.writeDigitRaw(1, data[1]);
    display.writeDigitRaw(3, data[2]);
    display.writeDigitRaw(4, data[3]);
    if (renderColon) {
        display.writeDigitRaw(2,
    		   (alarmArmed == 1 ? 0x04 : (alarmArmed == 2 ? 0x08 : 0)) | // Set alarm 1, 2 or none LED's
    		   (((second() % 2) == 0) ? 0 : 0x02)                      | // Blink colon LED's every second (even seconds is off odd is on)
    		      	    (timeReceived ? 0 : 0x10)                        // If we're waiting for a time update show degrees LED
			);
    } else {
        display.writeDigitRaw(2, data[4]);
    }

    display.writeDisplay();
#endif 
}

void switchRadio(bool bOn, bool bAck)
{
     // Check if radio or buzzer needs to be switched on
     send(msgRadioOn.set(bOn ? "1" : "0"), bAck);
     digitalWrite(PIN_RADIO, bOn ? RADIO_RELAY_ON : RADIO_RELAY_OFF);
     // Write some debug info
     Serial.print(F("Sw radio msg: "));
     Serial.println(bOn ? "on" : "off");

}

void wakeupLight(int setLevel) {
    if (alarmArmed && useLight) {  // No light or 
          // Calculate the wake time in total minutes (make sure we cover mid night-passing)
    	    int wakeTime    = ((snoozeHour * 60) + snoozeMinute + ((snoozeHour == 00 && snoozeMinute < WAKEUP_LIGHT_TIME) ? (24*60) : 0)) * 60;
          int currentTime = (hour() * 3600) +  (minute() * 60) + second();
          int secondsTillWakeUp = wakeTime - currentTime;
          byte newLevel = setLevel;

          if (setLevel != RENDER_VALUE) {
        	  secondsTillWakeUp = ((101 - setLevel) * WAKEUP_LIGHT_TIME * 60) / 100;
          } else {
        	  newLevel = (byte)(100 - ((secondsTillWakeUp * 100 ) / (WAKEUP_LIGHT_TIME * 60)));
        	  static byte lastLevel = 0xFF;
        	  if (lastLevel != newLevel) { // Tell GW if level was changed
        		  send(msgLightLevel.set(newLevel), true);
        		  lastLevel = newLevel;
        	  }
          }

          // Yes time to render light start first with red
          int red = fscale( 0, (WAKEUP_LIGHT_TIME * 60), MAX_RED, 0, secondsTillWakeUp, 4);
          // After half time add green light as well to make yellow
          int green = ((secondsTillWakeUp <= (WAKEUP_LIGHT_TIME * 30)) ? fscale( 0, (WAKEUP_LIGHT_TIME * 30), MAX_GREEN, 0, secondsTillWakeUp, 4) : 0);
          
          analogWrite(PIN_LIGHT_RED, red); 
          analogWrite(PIN_LIGHT_GREEN, green);

          Serial.print("Wake up light levels ");
          Serial.print(secondsTillWakeUp);
          Serial.print("s [");
          Serial.print(newLevel);
          Serial.print("%] -> R: ");
          Serial.print(red);
          Serial.print("  G: ");
          Serial.println(green);
    } 
}

void setWakeUpLight(bool bOn) {
    byte level = ((bOn == LIGHT_ON) ? MAX_RED : 0);
    analogWrite(PIN_LIGHT_RED, level); 

    level = ((bOn == LIGHT_ON) ? MAX_GREEN : 0);
    analogWrite(PIN_LIGHT_GREEN, level); 

    send(msgLightStatus.set(((bOn == LIGHT_ON) ? "1" : "0")), true);
    wait(100);
    send(msgLightLevel.set(((bOn == LIGHT_ON) ? 100 : 0)), true);
}

// This is called when a new time value was received
void receiveTime(unsigned long controllerTime) {
   // OK, set incoming time
   Serial.print(F("Time val rcvd: "));
   Serial.println(controllerTime);
   setTime(controllerTime); 
   timeReceived = true;
}

void setState(int newstate) {
	const char * states[] = {
			   "TIME",
			   "SET_ALARM",
			   "ALARM_ON",
			   "ALARM_UPCOMING",
			   "SNOOZING",
			   "PLAY_MUSIC",
			   "SHOW_TEMPERATUE",
			   "MANUAL_TRIGGER" };
	Serial.print("-- State from: STATE_");
	Serial.print(states[state]);
	Serial.print(" to STATE_");
	Serial.print(states[newstate]);
	Serial.println(" -- ");
	state = newstate;
}

void receive(const MyMessage &message) {
   if (message.isAck()) {
        Serial.println(F("GW ack"));
   }
   switch (message.type) {
      case V_LIGHT:
         switch (message.sensor) {
             case NODE_RADIO:
                 if (!message.isAck()) {
                    if (message.getBool()) {
                        // Change radio state external trigger -> ACK as regular Alarm
                        Serial.println(F("--- ALARM Ext. trigger ---"));
                        setState(STATE_ALARM_ON);
                        send(msgAlarmTripped.set("1"), true);
                        wait(100);
                        switchRadio(RADIO_ON, NO_ACK);   
                        if (useLight) {
                            setWakeUpLight(LIGHT_ON);
                        }
                        // Stop Alarm after 1 hour
                        snoozeHour   = (hour() + 1) % 24;
                        snoozeMinute = minute();
                    } else {
                        Serial.println(F("Turn off alrm"));
                        setState(STATE_TIME);
                        send(msgAlarmTripped.set("0"), true);
                        wait(100);
                        switchRadio(RADIO_OFF, NO_ACK);
                        setWakeUpLight(LIGHT_OFF);
                    }
                 }
                 break;

             case NODE_LIGHT:
            	 if (!message.isAck()) {
                 	 switch (state) {
                		case STATE_TIME:
                		case STATE_SET_ALARM:
                		case STATE_PLAY_MUSIC:
                		case STATE_SHOW_TEMPERATUE:
                			if (message.getBool()) {
                				// External trigger for starting wakeup light
                				setState(STATE_ALARM_UPCOMING);
                				snoozeMinute = (minute() + WAKEUP_LIGHT_TIME) % 60;
                				snoozeHour   = (hour() + (snoozeMinute < minute() ? 1 : 0)) % 24;
                			}
                		    break;

                		case STATE_ALARM_UPCOMING:
                			if (!message.getBool()) {
                				// Stop wakeup light procedure
                				setWakeUpLight(LIGHT_OFF);
                				setState(STATE_TIME);
                			}
                			break;

                		case STATE_SNOOZING:
                		case STATE_ALARM_ON:
                		case STATE_MANUAL_TRIGGER:
                			if (!message.getBool()) {
                				setWakeUpLight(LIGHT_OFF);
                			}
                			break;
                	 }
                 }
                 break;
         }
         break;
   
     case V_TRIPPED:
    	if (state == STATE_MANUAL_TRIGGER) {
    		// ACK for manual trigger let user know, by showing ON as well
    		if (message.getBool()) {
				man_data[2] = alon_data[2];
				man_data[3] = alon_data[3];
				setSegments(man_data, USE_DATA_4_COLON);
    		}
    	} else {
    		// Do not switch off when radio is playing music on manual request AND when external trigger switches off ALSO
    		// Do not switch on radio when manual trigger switches on
			if (state != STATE_PLAY_MUSIC) {
				digitalWrite(PIN_RADIO, message.getBool() ? RADIO_RELAY_ON : RADIO_RELAY_OFF);
			}
    	}
        // Write some debug info
        Serial.print(F("Alarm trip msg: "));
        Serial.println(message.getBool() ? "ALARM" : "rest");
        break;
   
     case V_ARMED:
        alarmArmed = message.getBool() ? (alarmArmed == 1 ? 2 : 1) : 0;
        if (state == STATE_TIME || state == STATE_PLAY_MUSIC || state == STATE_ALARM_UPCOMING) {
              setSegments(time_data, RENDER_COLON);
        } 
        saveState(EEPROM_ALARM_ARMED, alarmArmed);
        if (alarmArmed == 0) {
            // Make sure the lights are off         
            setWakeUpLight(LIGHT_OFF);
        }
        // Write some debug info
        Serial.print(F("Alarm clock msg: "));
        Serial.println(alarmArmed == 1 ? "Armed 1" : (alarmArmed == 2 ? "Armed 2" : "Bypass"));
        break;
        
     case V_TEMP:
        // Write some debug info
        Serial.print(F("Temp received: "));
        Serial.println(message.getFloat());
        tempToArray(temp_data, message.getFloat(), tempsensors_cnt);
        if (state == STATE_SHOW_TEMPERATUE) {
           setSegments(temp_data, USE_DATA_4_COLON);
        }
        break;
        
     case V_PERCENTAGE:
		if (!message.isAck()) {
			 switch (state) {
				case STATE_TIME:
				case STATE_SET_ALARM:
				case STATE_PLAY_MUSIC:
				case STATE_SHOW_TEMPERATUE:
					if (message.getByte() == 100) {
						if (alarmArmed) { // Wake procedure not started
							Serial.println(F("Ext trig str wakeup lght"));
							setState(STATE_ALARM_UPCOMING);
							snoozeMinute = (minute() + WAKEUP_LIGHT_TIME) % 60;
							snoozeHour   = (hour() + (snoozeMinute < minute() ? 1 : 0)) % 24;
						}
						if (alarmArmed == 0 || !useLight) { // Not armed or no light report back
							send(msgLightLevel.set(0), true);
						}
					} else {
						wakeupLight(message.getByte());
					}
					break;

				case STATE_ALARM_UPCOMING:
					if (message.getByte() == 0) {
						Serial.println(F("Stop wakeup light"));
						setWakeUpLight(LIGHT_OFF);
						setState(STATE_TIME);
					} else {
						wakeupLight(message.getByte());
					}
					break;

				case STATE_SNOOZING:
				case STATE_ALARM_ON:
				case STATE_MANUAL_TRIGGER:
					wakeupLight(message.getByte());
					break;
			 }
		 }
		 break;

     default:
        Serial.print(F("Unk msg: "));
        Serial.print(message.type);
        Serial.print(F(" pl: "));
        Serial.println(message.getString());
   }
}


/* fscale
 Floating Point Autoscale Function V0.1
 Paul Badger 2007
 Modified from code by Greg Shakar

 This function will scale one set of floating point numbers (range) to another set of floating point numbers (range)
 It has a "curve" parameter so that it can be made to favor either the end of the output. (Logarithmic mapping)

 It takes 6 parameters

 originalMin - the minimum value of the original range - this MUST be less than origninalMax
 originalMax - the maximum value of the original range - this MUST be greater than orginalMin

 newBegin - the end of the new range which maps to orginalMin - it can be smaller, or larger, than newEnd, to facilitate inverting the ranges
 newEnd - the end of the new range which maps to originalMax  - it can be larger, or smaller, than newBegin, to facilitate inverting the ranges

 inputValue - the variable for input that will mapped to the given ranges, this variable is constrained to originaMin <= inputValue <= originalMax
 curve - curve is the curve which can be made to favor either end of the output scale in the mapping. Parameters are from -10 to 10 with 0 being
          a linear mapping (which basically takes curve out of the equation)
 */
float fscale( float originalMin, float originalMax, float newBegin, float newEnd, float inputValue, float curve) {
  float OriginalRange = 0;
  float NewRange = 0;
  float zeroRefCurVal = 0;
  float normalizedCurVal = 0;
  float rangedValue = 0;
  boolean invFlag = 0;

  // condition curve parameter
  // limit range
  if (curve > 10) curve = 10;
  if (curve < -10) curve = -10;

  curve = (curve * -.1) ; // - invert and scale - this seems more intuitive - positive numbers give more weight to high end on output
  curve = pow(10, curve); // convert linear scale into logarithmic exponent for other pow function

  // Check for out of range inputValues
  if (inputValue < originalMin) {
    inputValue = originalMin;
  }
  if (inputValue > originalMax) {
    inputValue = originalMax;
  }

  // Zero Reference the values
  OriginalRange = originalMax - originalMin;

  if (newEnd > newBegin){ 
    NewRange = newEnd - newBegin;
  } else {
    NewRange = newBegin - newEnd; 
    invFlag = 1;
  }
  zeroRefCurVal = inputValue - originalMin;
  normalizedCurVal  =  zeroRefCurVal / OriginalRange;   // normalize to 0 - 1 float

  // Check for originalMin > originalMax  - the math for all other cases i.e. negative numbers seems to work out fine 
  if (originalMin > originalMax ) {
    return 0;
  }

  if (invFlag == 0){
    rangedValue =  (pow(normalizedCurVal, curve) * NewRange) + newBegin;
  } else  {   // invert the ranges
    rangedValue =  newBegin - (pow(normalizedCurVal, curve) * NewRange); 
  }
  return rangedValue;
}
